/*
 * MRustC - Rust Compiler
 * - By John Hodge (Mutabah/thePowersGang)
 *
 * expand/macro_rules.cpp
 * - Top-level handling of macro_rules! macros
 *  > macro_rules! dispatch handler
 *  > #[macro_use]
 *  > #[macro_export]
 *  > #[macro_reexport]
 */
#include <synext.hpp>
#include "../ast/expr.hpp"
#include "../ast/ast.hpp"
#include "../parse/common.hpp"
#include "../parse/ttstream.hpp"
#include <ast/crate.hpp>
#include <macro_rules/macro_rules.hpp>
#include <hir/hir.hpp>  // for HIR::Crate

class CMacroRulesExpander:
    public ExpandProcMacro
{
    ::std::unique_ptr<TokenStream> expand(const Span& sp, const ::AST::Crate& crate, const TokenTree& tt, AST::Module& mod) override
    {
        ERROR(sp, E0000, "macro_rules! requires an identifier" );
    }
    ::std::unique_ptr<TokenStream> expand_ident(const Span& sp, const ::AST::Crate& crate, const RcString& ident, const TokenTree& tt, AST::Module& mod) override
    {
        DEBUG("Parsing macro_rules! " << ident);
        TTStream    lex(sp, ParseState(crate.m_edition), tt);
        auto mac = Parse_MacroRules(lex);
        DEBUG("macro_rules! " << mod.path() + ident << " " << &*mac);
        mod.add_macro( false, ident, mv$(mac) );

        return ::std::unique_ptr<TokenStream>( new TTStreamO(sp, ParseState(crate.m_edition), TokenTree()) );
    }
};

class CMacroUseHandler:
    public ExpandDecorator
{
    AttrStage stage() const override { return AttrStage::Post; }

    void handle(const Span& sp, const AST::Attribute& mi, ::AST::Crate& crate, const AST::Path& path, AST::Module& mod, slice<const AST::Attribute> attrs, AST::Item& i) const override
    {
        TRACE_FUNCTION_F("[CMacroUseHandler] path=" << path);

        std::vector<RcString>   filter;
        std::vector<bool>   filters_used;
        if( mi.has_sub_items() )
        {
            filter.reserve(mi.items().size());
            filters_used.resize(mi.items().size());
            for( const auto& si : mi.items() )
            {
                const auto& name = si.name().as_trivial();
                filter.push_back(name);
            }
        }
        auto filter_valid = [&](RcString name)->bool {
            if( filter.empty() ) {
                return true;
            }
            auto it = std::find(filter.begin(), filter.end(), name);
            if(it != filter.end()) {
                auto i = it - filter.begin();
                filters_used[i] = true;
                return true;
            }
            else {
                return false;
            }
            };

        if(i.is_None()) {
            // Just ignore
        }
        else if(const auto* ec_item = i.opt_Crate())
        {
            const auto& ec = crate.m_extern_crates.at(ec_item->name.c_str());

            for(const auto& name : ec.m_hir->m_exported_macro_names)
            {
                if( !filter_valid(name) )
                {
                    continue;
                }
                ASSERT_BUG(sp, ec.m_hir->m_root_module.m_macro_items.count(name) == 1, "Macro `" << name << "` missing from crate " << ec.m_name);
                const auto* e = &*ec.m_hir->m_root_module.m_macro_items.at(name);
                if( !e->publicity.is_global() )
                {
                    continue ;
                }

                if( const auto* imp = e->ent.opt_Import() )
                {
                    const ::HIR::Module* mod = &crate.m_extern_crates.at(imp->path.m_crate_name).m_hir->m_root_module;
                    assert(imp->path.m_components.size() > 0);
                    for(size_t i = 0; i < imp->path.m_components.size() - 1; i ++)
                    {
                        const auto& cname = imp->path.m_components[i];
                        auto it = mod->m_mod_items.find(cname);
                        ASSERT_BUG(sp, it != mod->m_mod_items.end(), "Component " << i << " of " << imp->path << " not found");
                        const auto& se = (*it).second->ent;
                        ASSERT_BUG(sp, se.is_Module(), "Component " << i << " of " << imp->path << " not a module - " << se.tag_str());
                        mod = &se.as_Module();
                    }

                    e = &*mod->m_macro_items.at(name);
                    ASSERT_BUG(sp, !e->ent.is_Import(), "Recursive import");
                }

                TU_MATCH_HDRA( (e->ent), { )
                TU_ARMA(Import, imp) {
                    assert(false);
                    }
                TU_ARMA(MacroRules, mac_ptr) {
                    DEBUG("Imported " << name << "!");
                    mod.add_macro_import( name, *mac_ptr );
                    }
                TU_ARMA(ProcMacro, p) {
                    DEBUG("Imported " << name << "! (proc macro)");
                    auto mi = AST::Module::MacroImport{ false, p.path.m_components.back(), p.path.m_components, nullptr };
                    mi.path.insert(mi.path.begin(), p.path.m_crate_name);
                    mod.m_macro_imports.push_back(mv$(mi));
                    }
                }
            }
        }
        else if( const auto* submod_p = i.opt_Module() )
        {
            const auto& submod = *submod_p;
            for( const auto& mr : submod.macros() )
            {
                if( !filter_valid(mr.name) )
                {
                    continue;
                }
                DEBUG("Imported " << mr.name);
                mod.add_macro_import( mr.name, *mr.data );
            }
            for( const auto& mri : submod.macro_imports_res() )
            {
                if( !filter_valid(mri.name) )
                {
                    continue;
                }
                DEBUG("Imported " << mri.name << " (propagate)");
                mod.add_macro_import( mri.name, *mri.data );
            }
        }
        else {
            ERROR(sp, E0000, "Use of #[macro_use] on non-module/crate - " << i.tag_str());
        }

        for(size_t i = 0; i < filter.size(); i ++)
        {
            if( !filters_used[i] ) {
                ERROR(sp, E0000, "Couldn't find macro " << filter[i]);
            }
        }
    }

};

class CMacroExportHandler:
    public ExpandDecorator
{
    AttrStage stage() const override { return AttrStage::Post; }

    void handle(const Span& sp, const AST::Attribute& mi, ::AST::Crate& crate, const AST::Path& path, AST::Module& mod, slice<const AST::Attribute> attrs, AST::Item& i) const override
    {
        if( i.is_None() ) {
        }
        // If on a `use` it's for a #[rustc_builtin_macro]
        else if( const auto* u = i.opt_Use() )
        {
            if( u->entries.size() == 1
                && u->entries.back().path.is_absolute()
                && u->entries.back().path.m_class.as_Absolute().crate == CRATE_BUILTINS
                && u->entries.back().path.m_class.as_Absolute().nodes.size() == 1
                )
                ;
            else
                ERROR(sp, E0000, "Use of #[macro_export] on non-macro - " << i.tag_str());
            const auto& p = u->entries.back().path.m_class.as_Absolute();
            const auto& name = p.nodes.front().name();
            AST::Module::MacroImport    mi;
            mi.is_pub = true;
            mi.macro_ptr = nullptr;
            mi.name = u->entries.front().name;
            mi.path.push_back(p.crate);
            mi.path.push_back(name);
            crate.m_root_module.m_macro_imports.push_back(mv$(mi));

            crate.m_root_module.add_item(sp, true, name, i.clone(), {});
        }
        else if( i.is_MacroInv() ) {
            const auto& mac = i.as_MacroInv();
            if( !(mac.path().is_trivial() && mac.path().as_trivial() == "macro_rules") ) {
                ERROR(sp, E0000, "#[macro_export] is only valid on macro_rules!");
            }
            const auto& name = mac.input_ident();

            // Tag the macro in the module for crate export
            auto it = ::std::find_if( mod.macros().begin(), mod.macros().end(), [&](const auto& x){ return x.name == name; } );
            ASSERT_BUG(sp, it != mod.macros().end(), "Macro '" << name << "' not defined in this module");
            it->data->m_exported = true;
            DEBUG("- Export macro " << name << "!");
        }
        else {
            ERROR(sp, E0000, "Use of #[macro_export] on non-macro - " << i.tag_str());
        }
    }
};

class CMacroReexportHandler:
    public ExpandDecorator
{
    AttrStage stage() const override { return AttrStage::Post; }
    void handle(const Span& sp, const AST::Attribute& mi, ::AST::Crate& crate, const AST::Path& path, AST::Module& mod, slice<const AST::Attribute> attrs, AST::Item& i) const override
    {
        if( !i.is_Crate() ) {
            ERROR(sp, E0000, "Use of #[macro_reexport] on non-crate - " << i.tag_str());
        }

        const auto& crate_name = i.as_Crate().name;
        auto& ext_crate = *crate.m_extern_crates.at(crate_name.c_str()).m_hir;

        if( mi.has_sub_items() )
        {
            for( const auto& si : mi.items() )
            {
                if( !si.name().is_trivial() )
                    ERROR(sp, E0000, "macro_reexport of non-trivial name - " << si.name());
                const auto& name = si.name().as_trivial();
                auto it = ::std::find(ext_crate.m_exported_macro_names.begin(), ext_crate.m_exported_macro_names.end(), name);
                if( it == ext_crate.m_exported_macro_names.end() )
                    ERROR(sp, E0000, "Could not find macro " << name << "! in crate " << crate_name);
                // TODO: Do this differently.
                ext_crate.m_root_module.m_macro_items.at(name)->ent.as_MacroRules()->m_exported = true;
            }
        }
        else
        {
            ERROR(sp, E0000, "#[macro_reexport] requires a list of macros");
        }
    }
};

class CBuiltinMacroHandler:
    public ExpandDecorator
{
    AttrStage stage() const override { return AttrStage::Pre; }
    void handle(const Span& sp, const AST::Attribute& mi, ::AST::Crate& crate, const AST::Path& path, AST::Module& mod, slice<const AST::Attribute> attrs, AST::Item& i) const override
    {
        RcString    name;
        if(i.is_MacroInv()) {
            const auto& e = i.as_MacroInv();
            if( !(e.path().is_trivial() && e.path().as_trivial() == "macro_rules") ) {
                ERROR(sp, E0000, "Use of #[rustc_builtin_macro] on macro other than macro_rules! - " << i.tag_str());
            }
            name = e.input_ident();
        }
        else if(i.is_Macro()) {
            name = path.nodes().back().name();
        }
        else {
            ERROR(sp, E0000, "Use of #[rustc_builtin_macro] on non-macro - " << i.tag_str());
        }

        AST::UseItem    ui;
        ui.entries.push_back(AST::UseItem::Ent { });
        ui.entries.back().name = name;
        ui.entries.back().path = AST::Path(CRATE_BUILTINS, { name });
        DEBUG("Convert macro_rules tagged #[rustc_builtin_macro] with use");
        i = AST::Item::make_Use(mv$(ui));
    }
};

STATIC_MACRO("macro_rules", CMacroRulesExpander);
STATIC_DECORATOR("macro_use", CMacroUseHandler);
STATIC_DECORATOR("macro_export", CMacroExportHandler);
STATIC_DECORATOR("macro_reexport", CMacroReexportHandler);
STATIC_DECORATOR("rustc_builtin_macro", CBuiltinMacroHandler);

